import json
import requests
import odoo

from odoo.tests.common import TransactionCase, tagged
from odoo.addons.base_rest.tests.common import BaseRestCase


HOST = "localhost"
PORT = odoo.tools.config["http_port"]


class TestCRMLeadServiceRestCase(BaseRestCase):
    def setUp(self):
        super().setUp()
        self.api_key_test = self.env.ref("ce.auth_api_key_platform_admin_demo")
        self.AuthApiKey = self.env["auth.api.key"]
        self.session = requests.Session()

    @tagged('endpoints')
    def test_route_right_create(self):
        url = "/api/crm-lead"
        data = {
            "partner_name": "Aida Sanahuja",
            "email_from": "testmail@hotmail.com",
            "phone": "641708221",
            "odoo_company_id": 1,
            "source_id": 1,
        }

        response = self.http_post(url, data=data)
        self.assertEquals(response.status_code, 200)

        content = json.loads(response.content.decode("utf-8"))
        self.assertIn("id", content)

        crm_lead = self.env["crm.lead"].browse(content["id"])
        self.assertEquals(crm_lead.partner_name, data["partner_name"])
        self.assertEquals(crm_lead.phone, data["phone"])
        self.assertEquals(crm_lead.email_from, data["email_from"])
        self.assertEquals(crm_lead.company_id.id, data["odoo_company_id"])
        self.assertEquals(crm_lead.source_id.id, data["source_id"])

    def _add_api_key(self, headers):
        key_dict = {"API-KEY": self.api_key_test.key}
        if headers:
            headers.update(key_dict)
        else:
            headers = key_dict
        return headers

    def http_post(self, url, data, headers=None):
        headers = self._add_api_key(headers)
        if url.startswith("/"):
            url = "http://{}:{}{}".format(HOST, PORT, url)
            return self.session.post(url, json=data, headers=headers)



from odoo import fields, models, api

class ResUsersRoleLine(models.Model):
    _inherit = 'res.users.role.line'

    company_id = fields.Many2one(
        string='Companyia',
        comodel_name='res.company',
        default=lambda self: self.env["res.company"]._company_default_get()
    )


            
        




from odoo import fields, models, api


class KeycloakCreateWiz(models.TransientModel):
    """Export Odoo users to Keycloak.

    Usually Keycloak is already populated w/ your users base.
    Many times this will come via LDAP, AD, pick yours.

    Still, you might need to push some users to Keycloak on demand,
    maybe just for testing.

    If you need this, this is the wizard for you ;)
    """

    _inherit = 'auth.keycloak.create.wiz'

    def _create_user_values(self, odoo_user):
        """Prepare Keycloak values for given Odoo user."""

        values = super(KeycloakCreateWiz,self)._create_user_values(odoo_user)

        if self._context.get('kc_user_creation_vals'):
            values.update(self._context.get('kc_user_creation_vals'))

        return values



from odoo import api, models, fields, _
from odoo.exceptions import UserError
import logging
logger = logging.getLogger(__name__)


class ResUsers(models.Model):
    _inherit = 'res.users'

    # CAUTION: we are using here as keys, the 3 Keycloak role names:
    # ['role_ce_member', 'role_ce_admin', 'role_platform_admin']
    # Also those 3 names are the ones allowed to be used to be assigned to the user 'role' param in
    # any API call between Odoo and external services as CE-PublicWeb or CE-VirtualOffice
    USER_CE_ROLE_NAMES_SELECTION = [
        ('role_ce_member',_('CE Member')),
        ('role_ce_admin',_('CE Administrator')),
        ('role_platform_admin',_('Platform Administrator'))]

    ce_role = fields.Selection(USER_CE_ROLE_NAMES_SELECTION, string='CE Role', compute='_compute_ce_role')

    @api.depends('role_ids')
    def _compute_ce_role(self):

        role_ce_member_id = self.env['ir.model.data'].get_object_reference('ce','role_ce_member')[1]
        role_ce_admin_id = self.env['ir.model.data'].get_object_reference('ce','role_ce_admin')[1]
        role_platform_admin_id =  self.env['ir.model.data'].get_object_reference('ce','role_platform_admin')[1]

        for user in self:
            ce_role_ids = [r.id for r in user.role_ids if r.id in [role_ce_member_id, role_ce_admin_id, role_platform_admin_id]]

            role_value=None

            if role_platform_admin_id in ce_role_ids:
                role_value = 'role_platform_admin'
            elif role_ce_admin_id in ce_role_ids:
                role_value = 'role_ce_admin'
            elif role_ce_member_id in ce_role_ids:
                role_value = 'role_ce_member'

            user.ce_role = role_value

    @api.multi
    def push_new_user_to_keyckoack(self):
        self.ensure_one()

        ICPSudo = self.env['ir.config_parameter'].sudo()
        ce_admin_provider = self.company_id.ce_admin_key_cloak_provider_id
        ce_login_provider = self.company_id.auth_ce_key_cloak_provider_id

        if not ce_admin_provider:
            raise UserError(
                _("Unable to get the 'CE admin' provider_id related to tha current company when triying to push new user to KC."))

        wiz_vals = {
            'provider_id': ce_admin_provider.id,
            'user_ids': [(6, 0, [self.id])],
            'endpoint': ce_admin_provider.users_endpoint,
            'user': ce_admin_provider.superuser,
            'pwd': ce_admin_provider.superuser_pwd,
            'login_match_key': 'username:login'
        }

        ck_user_group_mapped_to_odoo_group_ce_member = ICPSudo.get_param(
            'ce.ck_user_group_mapped_to_odoo_group_ce_member')
        kc_user_additional_vals = {
            'attributes': {'lang': [self.lang]},
            'groups': [ck_user_group_mapped_to_odoo_group_ce_member],
            'enabled': True,
            'credentials': [
                {'type': 'password', 'value': 'w8P=FL_W', 'temporary': False}],
        }
        self = self.with_context(kc_user_creation_vals=kc_user_additional_vals)
        wiz = self.env['auth.keycloak.create.wiz'].create(wiz_vals)
        wiz.button_create_user()

        # after call the button_create_user() we need to re-set the user.oauth_provider_id
        # in order to ensure that it remains = ce_login_provider
        self.update({
            'oauth_provider_id': ce_login_provider,
        })

    @api.multi
    def update_user_data_to_keyckoack(self, kc_data_to_update=[]):
        """Proceed to update the related Keyckoak data for current user (=Self)

        :param kc_data_to_update: <list of strings> list of KeyCloak user data to update.
            If empty list proceed to update all the KC allowed updatable fields (see list: _KEYCLOAK_UPDATABLE_USER_DATA)
        """
        self.ensure_one()
        ce_admin_provider = self.company_id.ce_admin_key_cloak_provider_id

        if not ce_admin_provider:
            raise UserError(
                _("Unable to get the 'CE admin' provider_id related to tha current company when triying to push new user to KC."))

        if not self.oauth_uid:
            raise UserError(
                _("The Odoo user {} (Odoo id: {}) do not have any mapped Keycloak User ID.").format(self.login, self.id))

        wiz_vals = {
            'provider_id': ce_admin_provider.id,
            'endpoint': ce_admin_provider.users_endpoint,
            'user': ce_admin_provider.superuser,
            'pwd': ce_admin_provider.superuser_pwd,
            'login_match_key': 'username:login'
        }
        kc_wiz = self.env['auth.keycloak.sync.wiz'].create(wiz_vals)
        kc_wiz._validate_setup()

        token = kc_wiz._get_token()

        to_up_data_dict = kc_wiz._update_user_values(self, kc_data_to_update)

        if to_up_data_dict.get('groups'):

            new_group_allready_present_flag = False
            new_kc_group_name = to_up_data_dict.pop('groups')[0]

            # proceed to clear the existing groups
            group_ids_to_delete = []
            user_current_groups = kc_wiz._get_groups_from_user(token, self.oauth_uid) or []
            for group_dict in user_current_groups:
                if group_dict['name'] != new_kc_group_name:
                    group_ids_to_delete.append(group_dict['id'])
                else:
                    new_group_allready_present_flag = True
            if group_ids_to_delete:
                kc_wiz._delete_user_groups_to_user(token, self.oauth_uid, group_ids_to_delete)

            # proceed to add the new group
            if not new_group_allready_present_flag:
                # get the KC ID's for each KC group related to the user company-->realm
                new_kc_group_id = [g['id'] for g in self.company_id.get_kc_groups_data() if g['name'] == new_kc_group_name]
                new_kc_group_id = new_kc_group_id and new_kc_group_id[0]
                if not new_kc_group_id:
                    raise UserError(
                        _("Unable to get the Keycloak ID related to the Keycloak group '{}' for company/realm ID:{}").format(new_kc_group_name,self.company_id.id))

                kc_wiz._add_user_groups_to_user(token, self.oauth_uid, [new_kc_group_id])

        if to_up_data_dict:
            kc_wiz._update_user_by_id(token, self.oauth_uid, to_up_data_dict)

        # if all is OK we return the basic user KC representation
        return kc_wiz._get_user_by_id(token, self.oauth_uid)

    @api.model
    def ce_user_roles_mapping(self):

        ICPSudo = self.env['ir.config_parameter'].sudo()

        roles_map = {
            'role_ce_member':
                {'odoo_role_id': int(ICPSudo.get_param('ce.odoo_group_ce_member')),
                'kc_group_name': ICPSudo.get_param('ce.ck_user_group_mapped_to_odoo_group_ce_member'),
                'is_admin': False},
            'role_ce_admin':
                {'odoo_role_id': int(ICPSudo.get_param('ce.odoo_group_ce_admin')),
                'kc_group_name': ICPSudo.get_param('ce.ck_user_group_mapped_to_odoo_group_ce_admin'),
                'is_admin': True},
            'role_platform_admin':
                {'odoo_role_id': int(ICPSudo.get_param('ce.odoo_group_platform_admin')),
                'kc_group_name': ICPSudo.get_param('ce.ck_user_group_mapped_to_odoo_group_platform_admin'),
                'is_admin': True}
            }

        return roles_map



import requests

import odoo

from odoo.addons.base_rest.tests.common import BaseRestCase

HOST = "127.0.0.1"
PORT = odoo.tools.config["http_port"]


class BaseCERestCase(BaseRestCase):
    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super().setUpClass(*args, **kwargs)
        cls.AuthApiKey = cls.env["auth.api.key"]
        cls.api_key_test = cls.env.ref("ce.auth_api_key_platform_admin_demo")

    def setUp(self):
        super().setUp()
        self.session = requests.Session()

    def _add_api_key(self, headers):
        key_dict = {"API-KEY": self.api_key_test.key}
        if headers:
            headers.update(key_dict)
        else:
            headers = key_dict
        return headers

    def http_post(self, url, data, headers=None):
        headers = self._add_api_key(headers)
        if url.startswith("/"):
            url = "http://{}:{}{}".format(HOST, PORT, url)

        return self.session.post(url, json=data, headers=headers)



from odoo import api, models, fields, _
from odoo.exceptions import UserError, ValidationError

class SubscriptionRequest(models.Model):
    _inherit = 'subscription.request'

    gender = fields.Selection(
        [("male", _("Male")), ("female", _("Female")), ("not_binary", _("Not binary")),
        ("not_share", _("I prefer do not share it")), ("other", _("Other"))],
        string="Gender",
        readonly=True,
        states={"draft": [("readonly", False)]},
    )

    def get_journal(self):
        """Need to override in order to use in multicompany enviroment"""

        j = super().get_journal()

        if self.company_id.id != 1:
            if self.company_id.cooperator_journal:
                j = self.company_id.cooperator_journal
            else:
                raise UserError(_("You must set a cooperator journal on you company."))

        return j

    def get_required_field(self):
        required_fields =  super(SubscriptionRequest,self).get_required_field()
        if 'iban' in required_fields: required_fields.remove('iban')
        return required_fields

    @api.model
    def create(self, vals):
        if vals:
            vals['skip_iban_control'] = True
        subscription_request = super(SubscriptionRequest, self).create(vals)
        return subscription_request



import logging, json
from odoo.addons.base_rest import restapi
from werkzeug.exceptions import BadRequest, NotFound
from odoo.addons.base_rest.http import wrapJsonException
from odoo.addons.component.core import Component
from odoo import _
from . import schemas

_logger = logging.getLogger(__name__)

class MemberService(Component):
    _inherit = "base.rest.private_abstract_service"
    _name = "ce.member.services"
    _collection = "ce.services"
    _usage = "member"
    _description = """
        CE Member roles requests
    """

    @restapi.method(
        [(["/<string:keycloak_id>"], "GET")],
        output_param=restapi.CerberusValidator("_validator_return_get"),
        auth="api_key",
    )
    def get(self, _keycloak_id):
        user, partner, role, email = self._get_member_profile_objs(_keycloak_id)
        return self._to_dict(user, partner, role, email)
    
    def _validator_return_get(self):
        return schemas.S_MEMBER_PROFILE_RETURN_GET

    @restapi.method(
        [(["/<string:keycloak_id>"], "PUT")],
        input_param=restapi.CerberusValidator("_validator_update"),
        output_param=restapi.CerberusValidator("_validator_return_update"),
        auth="api_key",
    )
    def update(self, _keycloak_id, **params):
        _logger.info("Requested role update, user: {}".format(_keycloak_id))

        user, partner, actual_role, email = self._get_member_profile_objs(_keycloak_id)
        ce_roles_map = user.ce_user_roles_mapping()

        new_role = params['role']

        if new_role not in ce_roles_map.keys():
            valid_roles = [item[0] for item in self.env['res.users'].USER_CE_ROLE_NAMES_SELECTION]
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={
                    'message': _("The role '{}' is not a valid one, it must be one of: {}").format(
                        params['role'],
                        valid_roles),
                    'code': 500,
                    }
            )

        if new_role != actual_role:

            RoleLineSudo = self.env['res.users.role.line'].sudo()

            for role_line in user.role_line_ids:
                role_id = role_line.role_id.id
                if role_id in [r['odoo_role_id'] for r in ce_roles_map.values()]:
                    role_line.sudo().unlink()

            RoleLineSudo.create({'user_id': user.id, 'role_id': ce_roles_map[new_role]['odoo_role_id']})

            if not user.company_id.check_ce_has_admin():
                raise wrapJsonException(
                    BadRequest(),
                    include_description=False,
                    extra_info={
                        'message': _("Unable to change the member role to '{}', because the company would remain without any 'CE Admin' member").format(new_role),
                        'code': 500,
                        }
                )

            #also update KeyCloack GROUP acordingly throw API call
            try:
                user.update_user_data_to_keyckoack(['groups'])
            except Exception as ex:
                details = (ex and hasattr(ex, 'name') and " Details: {}".format(ex.name)) or ''
                raise wrapJsonException(
                    BadRequest(),
                    include_description=False,
                    extra_info={
                        'message': _("Unable to update the role in Keycloak for the related KC user ID: {}.{}").format(
                            user.oauth_uid,
                            details),
                        'code': 500,
                    })

        return self._to_dict(user, partner, new_role, email)

    def _prepare_create(self, params):
        """Prepare a writable dictionary of values"""
        return {
            "company_id": params["company_id"],
            "user_id": params["user_id"],
            "target_company_id": params["target_company_id"],
            "target_user_id": params["target_user_id"],
            "new_role": params["new_role"],
        }

    def _validator_update(self):
        return schemas.S_MEMBER_PROFILE_PUT

    def _validator_return_update(self):
        return schemas.S_MEMBER_PROFILE_RETURN_PUT
    

    def _get_member_profile_objs(self, _keycloak_id):
        user = self.env["res.users"].sudo().search([('oauth_uid','=',_keycloak_id)])
        if not user:
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={'message': _("No Odoo User found for KeyCloak user id %s") % _keycloak_id}
            )

        partner = user.partner_id or None
        if not partner:
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={'message': _("No Odoo Partner found for Odoo user with login username %s") % user.login}
            )

        role = user.ce_role or ''
        email = partner.email or False

        return user, partner, role, email


    @staticmethod
    def _to_dict(user, partner, role, email):
        return {
            'member':{
                "keycloak_id": user.oauth_uid,
                "name": user.display_name,
                "role": user.ce_role or "",
                "email": email,
            }
        }


import odoo

from .common import BaseCERestCase


class BaseCERestCaseAdmin(BaseCERestCase):
    @classmethod
    def setUpClass(cls, *args, **kwargs):
        # Skip parent class in super to avoid recreating api key
        super(BaseCERestCaseAdmin, cls).setUpClass(*args, **kwargs)
        AuthApiKey = cls.env["auth.api.key"]
        admin = cls.env.ref("base.user_admin")
        cls.api_key_test = AuthApiKey.create(
            {"name": "test-key", "key": "api-key", "user_id": admin.id}
        )
        cls.env = cls.env(context=dict(cls.env.context, tracking_disable=True))

    def http_get_without_auth(self):
        HOST = "127.0.0.1"
        PORT = odoo.tools.config["http_port"]
        url = "http://{}:{}{}".format(HOST, PORT, self.url)
        return self.session.get(url)



# Copyright 2018 Camptocamp SA
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)
from argparse import RawDescriptionHelpFormatter
from odoo import fields, models, api, _
from odoo.exceptions import UserError
import logging
import requests
try:
    from json.decoder import JSONDecodeError
except ImportError:
    # py2
    JSONDecodeError = ValueError

logger = logging.getLogger(__name__)

class KeycloakSyncMixin(models.AbstractModel):
    """Synchronize Keycloak users mixin."""

    _inherit = 'auth.keycloak.sync.mixin'

    _KEYCLOAK_UPDATABLE_USER_DATA = ['lang', 'groups', 'enabled', 'username', 'email', 'firstname_lastname', 'credentials']

    # MANAGE REALM DATA ______________________________________________________________

    def _get_realm_groups_data(self, token):
        """Get the JSON Representation of groups of a given REALM.
        It require the KC 'master' admin user have 'query-groups' Client Role assigned for the target realm
        It returns a list of dicts. ex:
        [{
            "id": "8ge163b3-6kc7-40ed-x069-3309eabbcbea",
            "name": "group1",
            "path": "/group1",
            "subGroups": []
        }] """
        target_endpoint = '{}/groups'.format(self.endpoint.replace('/users',''))

        logger.info('GET KEYCLOAK GROUPS DATA FOR REALM %s' % self.endpoint)
        headers = {
            'Authorization': 'Bearer %s' % token,
        }
        resp = requests.get(target_endpoint, headers=headers, json={})
        self._validate_response(resp, no_json=True)

        return resp.json()


    # MANAGE USER DATA ______________________________________________________________

    def _get_user_by_id(self, token, user_keycloak_id):
        """Get the JSON Representation of a given keycloak user ID"""

        target_endpoint = '{}/{}'.format(self.endpoint, user_keycloak_id)
        
        logger.info('GET KEYCLOAK USER DATA BY ID Calling %s' % self.endpoint)
        headers = {
            'Authorization': 'Bearer %s' % token,
        }
        resp = requests.get(target_endpoint, headers=headers, json={})
        self._validate_response(resp, no_json=True)

        return resp.json()

    def _get_groups_from_user(self, token, user_keycloak_id):
        """Get the GROUPS JSON Representation of a given keycloak user ID.
        It will retiurn a list of dicts, ex:
        [
            {'id': '72f8045b-984a-4773-bf59-461764c9d16e', 'name': 'ce_admins_group', 'path': '/ce_admins_group'}, 
            {'id': '59139b95-0066-45a6-9b0b-d8bdf4006f87', 'name': 'ce_members_group', 'path': '/ce_members_group'}
        ]
        """
        target_endpoint = '{}/{}/groups'.format(self.endpoint, user_keycloak_id)

        logger.info('GET KEYCLOAK USER GROUPS DATA BY ID Calling %s' % self.endpoint)
        headers = {
            'Authorization': 'Bearer %s' % token,
        }
        resp = requests.get(target_endpoint, headers=headers, json={})
        self._validate_response(resp, no_json=True)
        return resp.json()


    def _delete_user_groups_to_user(self, token, user_keycloak_id, group_ids):
        """Remove a given list of KC group_ids from a user."""

        headers = {
                'Authorization': 'Bearer %s' % token,
            }
        try:
            for group_id in group_ids:
                target_endpoint = '{}/{}/groups/{}'.format(self.endpoint, user_keycloak_id, group_id)
                logger.info('DELETE KEYCLOAK USER GROUP BY ID Calling %s' % self.endpoint)
                requests.delete(target_endpoint, headers=headers, json={})
        except:
            raise UserError(
                _("Error when calling Keycloak API to delete group on user: {}").format(target_endpoint))

        return True

    def _update_user_values(self, odoo_user, kc_data_to_update=[]):
        """Prepare Keycloak UPDATE values for given Odoo user."""

        values = {}

        if not kc_data_to_update:
            # update all the allowed KC user data
            kc_data_to_update = self._KEYCLOAK_UPDATABLE_USER_DATA

        #discart not allowed
        kc_data_to_update = [kc_field for kc_field in kc_data_to_update if kc_field in self._KEYCLOAK_UPDATABLE_USER_DATA]
        
        if 'lang' in kc_data_to_update:
            if values.get('attributes'):
                values['attributes'].update({'lang': odoo_user.lang})
            else:
                values.update({'attributes':{'lang': odoo_user.lang},})

        if 'groups' in kc_data_to_update:
            ce_roles_map = odoo_user.ce_user_roles_mapping()
            values['groups'] = [ce_roles_map[odoo_user.ce_role]['kc_group_name']]


        if 'enabled' in kc_data_to_update:
            values['enabled'] = odoo_user.active

        if 'username' in kc_data_to_update:
            values['username'] = odoo_user.login

        if 'email' in kc_data_to_update:
            values['email'] = odoo_user.email

        if 'firstname_lastname' in kc_data_to_update:
            if 'firstname' in odoo_user.partner_id:
                # partner_firstname installed
                values['firstName'] = odoo_user.partner_id.firstname
                values['lastName'] = odoo_user.partner_id.lastname
            else:
                values['firstName'], values['lastName'] = self._split_user_fullname(odoo_user)

        if 'credentials' in kc_data_to_update:
            pass
            # todo: implement here the logic needed to do the password update from Odoo to KC
            #values['credentials'] = [{'type': 'password', 'value': 'w8P=FL_W', 'temporary': False}]

        return values
    
    def _update_user_by_id(self, token, user_keycloak_id, to_up_data_dict):
        """Update a given keycloak user ID w/ given data."""

        target_endpoint = '{}/{}'.format(self.endpoint, user_keycloak_id)

        logger.info('UPDATE KEYCLOAK USER DATA BY ID Calling %s' % self.endpoint)
        headers = {
            'Authorization': 'Bearer %s' % token,
        }
        
        try:
            requests.put(target_endpoint, headers=headers, json=to_up_data_dict)
        except:
            raise UserError(
                _("Error when calling Keycloak API to update user: {}").format(target_endpoint))

        return True

    def _add_user_groups_to_user(self, token, user_keycloak_id, group_ids):
        """Assign(add) a given list of KC group_ids to a user."""

        headers = {
                'Authorization': 'Bearer %s' % token,
            }
        try:
            for group_id in group_ids:
                target_endpoint = '{}/{}/groups/{}'.format(self.endpoint, user_keycloak_id, group_id)
                logger.info('UPDATE KEYCLOAK USER GROUP BY ID Calling %s' % self.endpoint)
                requests.put(target_endpoint, headers=headers, json={})
        except:
            raise UserError(
                _("Error when calling Keycloak API to add group on user: {}").format(target_endpoint))

        return True




from odoo import fields, models, api

class OAuthProvider(models.Model):
    _inherit = 'auth.oauth.provider'

    company_id = fields.Many2one(
        string='Companyia',
        comodel_name='res.company',
        default=lambda self: self.env["res.company"]._company_default_get()
    )
    
    login_provider = fields.Boolean(string='For users login')

    @api.model
    def update_ce_oauth_providers(self,kc_url,kc_port):
        if any([kc_url,kc_port]):
            pass #todo: update the existing providers related to the ce module accordingly

            
        




from odoo import models, api, fields, _
from odoo.exceptions import UserError
from datetime import datetime


class CrmLead(models.Model):
    _inherit = 'crm.lead'

    # mapping between the xml_ids related to crm.lead.tag_ids and cm.filter xmlids
    _XMLID_MAPPING_LEADTAGS_CMFILTERS = [
        ('ce_tag_common_generation','ce_cm_filter_community_renewal_generation'),
        ('ce_tag_energy_efficiency','ce_cm_filter_energetic_eficiency'),
        ('ce_tag_sustainable_mobility','ce_cm_filter_sustainable_mobility'),
        ('ce_tag_citizen_education','ce_cm_filter_citizen_education'),
        ('ce_tag_thermal_energy','ce_cm_filter_thermical_energy'),
        ('ce_tag_collective_purchases','ce_cm_filter_collective_purchase'),
        ('ce_tag_renewable_energy','ce_cm_filter_renewal_energy_supply'),
        ('ce_tag_aggregate_demand','ce_cm_filter_demand_flexibility_and_aggregation')
    ]

    lang = fields.Char(string="Language")
    ce_tag_ids = fields.Many2many('crm.lead.tag', 'crm_lead_ce_tag_rel', 'lead_id', 'tag_id', string='CE Tags', help="CE Classify and analyze categories")

    community_company_id = fields.Many2one(
        string='Related Community',
        comodel_name='res.company',
        domain="[('coordinator','!=',True)]",
        help="Community related to this Lead"
    )

    @api.multi
    def _create_map_place_proposal(self):

        if not self.env.user.company_id.coordinator:
            raise UserError(
                _("Only users that belongs to the 'Coordinator' company can create new Map Places from Leads."))

        active_categ_id = self.env['ir.model.data'].get_object_reference('ce', 'ce_cm_place_category_active')[1]
        building_categ_id = self.env['ir.model.data'].get_object_reference('ce', 'ce_cm_place_category_building')[1]
        default_ce_map_id = self.env['ir.model.data'].get_object_reference('ce', 'ce_default_cm_map')[1]
        creation_ce_source_id = self.env['ir.model.data'].get_object_reference('ce', 'ce_source_creation_ce_proposal')[1]

        for lead in self:

            if self.env['crm.team'].search([('proposal_form_submission_id','=',lead.id), ('map_id','=',default_ce_map_id)]):
                raise UserError(
                    _("There is an allready existing Map Place related to this Lead: {}.").format(lead.name))

            if not lead.source_id or lead.source_id.id != creation_ce_source_id:
                raise UserError(
                    _("The Source {} of Lead {} do not allow the creation of Map Proposals").format(lead.source_id.name, lead.name))

            place_creation_data = {
                'name': lead.name,
                'map_id': default_ce_map_id,
                'team_type': 'map_place_proposal',
                'user_id': self.env.user.id,
                'proposal_form_submission_id': lead.id,
                'interaction_method': 'external_link',
                'external_link_target': '_top',
            }

            # read metadata key/value pairs
            m_dict = {m.key: m.value for m in lead.form_submission_metadata_ids}

            if m_dict.get('partner_legal_state',False) and m_dict['partner_legal_state']:
                if m_dict['partner_legal_state'] == 'active':
                    place_creation_data['place_category_id'] = active_categ_id
                else:
                    place_creation_data['place_category_id'] = building_categ_id
            else:
                raise UserError(
                    _("Unable to get the Category (mandatory map place field) from Lead: {}").format(lead.name))

            if m_dict.get('partner_latitude',False) and m_dict['partner_latitude']:
                place_creation_data['lat'] = m_dict['partner_latitude']
            else:
                raise UserError(
                    _("Unable to get the Latitude (mandatory map place field) from Lead: {}").format(lead.name))

            if m_dict.get('partner_longitude',False) and m_dict['partner_longitude']:
                place_creation_data['lng'] = m_dict['partner_longitude']
            else:
                raise UserError(
                    _("Unable to get the Longitude (mandatory map place field) from Lead: {}").format(lead.name))

            if m_dict.get('partner_map_place_form_url',False) and m_dict['partner_map_place_form_url']:
                place_creation_data['external_link_url'] = m_dict['partner_map_place_form_url']

            place_creation_data['address_txt'] = lead._get_address_txt() or None
            place_creation_data['filter_mids'] = [(6,0,lead._get_cmfilter_ids())]

            place = self.env['crm.team'].create(place_creation_data)

            # we need to do call those next 2 functions because the @api.onchange('map_id') defined on community_maps module
            # is not being called on crm_team.create(). TODO: review why it happens and fix it.
            place._get_slug_id()
            place._get_config_relations_attrs()
            place._build_presenter_metadata_ids()
            place.place_category_id = place_creation_data['place_category_id']

            place.message_subscribe([self.env.user.partner_id.id])

            pmnd_ids = [m for m in place.place_presenter_metadata_ids if m.key == 'p_description']
            description_pmnd_id = pmnd_ids and pmnd_ids[0] or None
            if description_pmnd_id and lead.description:
                description_pmnd_id.value = "<p class='m-2'>{}</p>".format(lead.description)

            # lead update
            lead.write({
                'team_id': place.id,
                'submission_type': 'place_proposal_submission',
            })

    @api.multi
    def _get_cmfilter_ids(self):
        self.ensure_one()

        md = self.env['ir.model.data']
        id_pairs = {}
        for pair in self._XMLID_MAPPING_LEADTAGS_CMFILTERS:
            id_pairs[md.get_object_reference('ce', pair[0])[1]] = md.get_object_reference('ce', pair[1])[1]

        return [id_pairs[t.id] for t in self.tag_ids if t.id in id_pairs]

    @api.multi
    def _get_address_txt(self):
        self.ensure_one()

        ret = ''
        meta_address_txt = [meta.value for meta in self.form_submission_metadata_ids if meta.key == 'partner_full_address']

        if self.street and (self.city or self.zip):
            ret ="{}{}. {}{}".format(
                self.street,
                (self.street2 and ' '+self.street2) or '',
                self.zip or '',
                (self.city and ' '+self.city) or '')
            if self.state_id:
                ret += ', {}'.format(self.state_id.name)
            if self.country_id:
                ret += '. {}'.format(self.country_id.name)
        elif meta_address_txt and meta_address_txt[0]:
            ret = meta_address_txt[0]

        return ret

    @api.multi
    def _build_community_company(self):

        if not self.env.user.company_id.coordinator:
            raise UserError(
                _("Only users that belongs to the 'Coordinator' company can create new Companies from Leads."))

        creation_ce_source_id = self.env['ir.model.data'].get_object_reference('ce', 'ce_source_creation_ce_proposal')[1]

        # build company for each Lead
        for lead in self:

            if not lead.source_id or lead.source_id.id != creation_ce_source_id:
                raise UserError(
                    _("The Source {} of Lead {} do not allow the creation of new Companies").format(lead.source_id.name, lead.name))

            if not lead.community_company_id:
                # Create the new company using very basic starting Data
                company = self.env['res.company'].create(lead._get_company_create_vals())

                # Update Lead & Map Place (if exist) fields accordingly
                lead.community_company_id = company.id
                if lead.team_id and lead.team_type == 'map_place_proposal':
                    lead.team_id.community_company_id = company.id

                # Do post creation specific CCEE tasks
                company._community_post_company_creation_tasks()

        # we need to do this commit before proceed to call KeyCloak API calls to build the related KC realms
        self._cr.commit()

        # build keyKloac realm for each new existing new company
        for lead in self:
            if lead.community_company_id:
                lead.community_company_id._create_keycloak_realm()
                lead.community_company_id._community_post_keycloak_creation_tasks()

    @api.multi
    def _get_company_create_vals(self):
        self.ensure_one()
        m_dict = {m.key: m.value for m in self.form_submission_metadata_ids}

        foundation_date = None
        if m_dict.get('partner_foundation_date',False) and m_dict['partner_foundation_date']:
            date_formats = ['%Y-%m-%d','%d-%m-%Y','%Y/%m/%d','%d/%m/%Y']
            for date_format in date_formats:
                try:
                    foundation_date = datetime.strptime(m_dict['partner_foundation_date'], date_format)
                except:
                    pass
            if not foundation_date:
                raise UserError(
                    _("The Foundation Date value {} have a non valid format. It must be: yyyy-mm-dd or dd-mm-yyyy or yyyy/mm/dd or dd/mm/yyyy").format(m_dict['partner_foundation_date']))

        initial_share_amount = 0.00
        if m_dict.get('partner_initial_share_amount',False) and m_dict['partner_initial_share_amount'] or None:
            try:
                initial_share_amount = float(m_dict['partner_initial_share_amount'])
            except:
                pass

        lang_id = None
        if m_dict.get('partner_language',False) and m_dict['partner_language'] or None:
            lang_id = self.env['res.lang'].search([('code','=',m_dict['partner_language'])],limit=1)


        create_vals = {
                'name': self.name,
                'street': self.street,
                'street2': self.street2,
                'city':self.city,
                'zip': self.zip,
                'state_id': self.state_id.id,
                'country_id': self.country_id.id,
                'website': self.website,
                'phone': self.phone,
                'email': self.email_from or (m_dict.get('contact_email',False) and m_dict['contact_email']) or None,
                'vat': m_dict.get('partner_vat',False) and m_dict['partner_vat'] or None,
                'social_twitter': m_dict.get('partner_twitter',False) and m_dict['partner_twitter'] or None,
                'social_facebook': m_dict.get('partner_facebook',False) and m_dict['partner_facebook'] or None,
                'social_instagram': m_dict.get('partner_instagram',False) and m_dict['partner_instagram'] or None,
                'social_telegram': m_dict.get('partner_telegram',False) and m_dict['partner_telegram'] or None,
                'create_user': True,
                'foundation_date': foundation_date,
                'initial_subscription_share_amount': initial_share_amount,
                'default_lang_id': lang_id and lang_id.id or None,
            }

        return create_vals


    @api.multi
    def _create_keycloak_realm(self):
        for lead in self:
            if not lead.community_company_id:
                raise UserError(
                    _("Unable to create the KeyCloack entities from Lead: {}, because it is not yet related to any Community company").format(lead.name))
            lead.community_company_id._create_keycloak_realm()


    @api.multi
    def _create_community_initial_users(self):
        for lead in self:
            pass


class CrmLeadTags(models.Model):
    _inherit = 'crm.lead.tag'

    tag_ext_id = fields.Char('ID Ext tag', compute='compute_ext_id_tag')

    def compute_ext_id_tag(self):
        for record in self:
            res = record.get_external_id()
            record.tag_ext_id = False
            if res.get(record.id):
                record.tag_ext_id = res.get(record.id)


from odoo import api, models, fields, _

class ResConfigSettings(models.TransientModel):
    _inherit = 'res.config.settings'

    # global parameters (saved on ir.config_parameter):
    ck_server_url = fields.Char(string="KeyCloak Server URL", config_parameter='ce.ck_server_url', placeholder="http://keycloak/")
    ck_server_port = fields.Char(string="KeyCloak Server port", config_parameter='ce.ck_server_port', placeholder="8080")
    ck_platform_oauth_provider_id = fields.Many2one('auth.oauth.provider', string='OAuth provider for Platform admin',
        config_parameter='ce.ck_platform_oauth_provider_id',domain="[('company_id','=',company_id)]")
    ck_user_group_mapped_to_odoo_group_ce_member = fields.Char(string="KeyCloak user group (ce_members)",
    config_parameter='ce.ck_user_group_mapped_to_odoo_group_ce_member')
    odoo_group_ce_member = fields.Many2one(
        'res.users.role', string="Odoo user group for ce_members", config_parameter='ce.odoo_group_ce_member')
    odoo_group_ce_admin = fields.Many2one(
        'res.users.role', string="Odoo user group for ce_admin", config_parameter='ce.odoo_group_ce_admin')
    odoo_group_platform_admin = fields.Many2one(
        'res.users.role', string="Odoo user group for platform_admin", config_parameter='ce.odoo_group_platform_admin')

    ck_user_group_mapped_to_odoo_group_ce_admin = fields.Char(string="KeyCloak user group (ce_admins)",
    config_parameter='ce.ck_user_group_mapped_to_odoo_group_ce_admin')
    ck_user_group_mapped_to_odoo_group_platform_admin = fields.Char(string="KeyCloak user_group (platform_admins)",
    config_parameter='ce.ck_user_group_mapped_to_odoo_group_platform_admin')

    # company dependent parameters (saved on res_company):
    kc_realm = fields.Char(string='KeyCloak realm name', related='company_id.kc_realm', readonly=False)
    ce_admin_key_cloak_provider_id = fields.Many2one('auth.oauth.provider',
        string='OAuth provider for CCEE admin', readonly=False, related='company_id.ce_admin_key_cloak_provider_id',
        domain="[('company_id','=',company_id)]")
    auth_ce_key_cloak_provider_id = fields.Many2one('auth.oauth.provider',
        string='OAuth provider for CCEE login', readonly=False, related='company_id.auth_ce_key_cloak_provider_id',
        domain="[('company_id','=',company_id)]")

    def set_values(self):
        ICPSudo = self.env['ir.config_parameter'].sudo()

        pre_ck_server_url = ICPSudo.get_param('ce.ck_server_url')
        pre_ck_server_port = ICPSudo.get_param('ce.ck_server_port')
        
        super(ResConfigSettings, self).set_values()
        
        ICPSudo.set_param('ce.ck_server_url', self.ck_server_url)
        ICPSudo.set_param('ce.ck_server_port', self.ck_server_port)
        ICPSudo.set_param('ce.ck_platform_oauth_provider_id', self.ck_platform_oauth_provider_id.id)
        ICPSudo.set_param('ce.ck_user_group_mapped_to_odoo_group_ce_member', self.ck_user_group_mapped_to_odoo_group_ce_member)
        ICPSudo.set_param('ce.ck_user_group_mapped_to_odoo_group_ce_admin', self.ck_user_group_mapped_to_odoo_group_ce_admin)
        ICPSudo.set_param('ce.ck_user_group_mapped_to_odoo_group_platform_admin', self.ck_user_group_mapped_to_odoo_group_platform_admin)

        self.env['auth.oauth.provider'].update_ce_oauth_providers(
            pre_ck_server_url != self.ck_server_url and self.ck_server_url or None,
            pre_ck_server_port != self.ck_server_port and  self.ck_server_port or None)

    @api.model
    def get_values(self):
        res = super(ResConfigSettings, self).get_values()
        ICPSudo = self.env['ir.config_parameter'].sudo()
        ck_server_url = ICPSudo.get_param('ce.ck_server_url')
        ck_server_port = ICPSudo.get_param('ce.ck_server_port')
        ck_platform_oauth_provider_id = ICPSudo.get_param('ce.ck_platform_oauth_provider_id')
        ck_user_group_mapped_to_odoo_group_ce_member = ICPSudo.get_param('ce.ck_user_group_mapped_to_odoo_group_ce_member')
        ck_user_group_mapped_to_odoo_group_ce_admin = ICPSudo.get_param('ce.ck_user_group_mapped_to_odoo_group_ce_admin')
        ck_user_group_mapped_to_odoo_group_platform_admin = ICPSudo.get_param('ce.ck_user_group_mapped_to_odoo_group_platform_admin')

        res.update(
            ck_server_url=ck_server_url,
            ck_server_port=ck_server_port,
            ck_platform_oauth_provider_id=int(ck_platform_oauth_provider_id),
            ck_user_group_mapped_to_odoo_group_ce_member=ck_user_group_mapped_to_odoo_group_ce_member,
            ck_user_group_mapped_to_odoo_group_ce_admin=ck_user_group_mapped_to_odoo_group_ce_admin,
            ck_user_group_mapped_to_odoo_group_platform_admin=ck_user_group_mapped_to_odoo_group_platform_admin,
        )

        return res


def boolean_validator(field, value, error):
    if value and value not in ["true", "false"]:
        error(field, "Must be a boolean value: true or false")

def ce_state_validator(field, value, error):
    if value and value not in ["active", "on_building"]:
        error(field, "Must be 'active' or 'on_building'")

S_CRM_LEAD_RETURN_CREATE = {
    "id": {"type": "integer"},
}

S_CRM_LEAD_CREATE = {
    "partner_name": {"type": "string"},
    "partner_email": {"type": "string"},
    "partner_phone": {"type": "string"},
    "partner_full_address": {"type": "string"},
    "partner_city": {"type": "string"},
    "partner_zip": {"type": "string"},
    "odoo_company_id": {"type": "integer"},
    "source_xml_id": {"type": "string"},
    "tag_ids": {
        "type": "list",
        "schema": {
            "type": "integer",
        }
    },
}

S_CRM_LEAD_CREATE_ALTA_CE = {
    "partner_name": {"type": "string", "required": True},
    "partner_description": {"type": "string", "required": True},
    "partner_full_address": {"type": "string", "required": True},
    "partner_zip": {"type": "string", "required": True},
    "partner_city": {"type": "string", "required": True},
    "partner_state": {"type": "string", "required": True},
    "partner_qty_members": {"type": "integer", "required": True},
    "partner_legal_state": {"type": "string",
        "check_with": ce_state_validator
    },
    "tag_ids": {
        "type": "list",
        "schema": {
            "type": "integer",
        },
        "required": True
    },
    "partner_foundation_date": {"type": "string"},
    "partner_vat": {"type": "string"},
    "partner_comments": {"type": "string"},
    "partner_firstname": {"type": "string", "required": True},
    "partner_lastname": {"type": "string", "required": True},
    "partner_email": {"type": "string", "required": True},
    "partner_phone": {"type": "string", "required": True},
    "contact2_firstname": {"type": "string"},
    "contact2_lastname": {"type": "string"},
    "contact2_email": {"type": "string"},
    "contact2_mobile": {"type": "string"},
    "odoo_company_id": {"type": "integer", "required": True},
    "source_xml_id": {"type": "string", "required": True},
    "partner_map_place_form_url":{"type": "string", "required": False},
    "partner_language": {"type": "string"},
}

S_PROFILE_RETURN_GET = {
    "profile": {
        "type": "dict",
        "schema": {
            "keycloak_id": {"type": "string"},
            "name": {"type": "string"},
            "surname": {"type": "string"},
            "birth_date": {"type": "string"},
            "gender": {"type": "string"},
            "vat": {"type": "string"},
            "contact_info": {
                "type": "dict",
                "schema": {
                    "email": {"type": "string"},
                    "phone": {"type": "string"},
                    "street": {"type": "string"},
                    "postal_code": {"type": "string"},
                    "city": {"type": "string"},
                    "state": {"type": "string"},
                    "country": {"type": "string"},
                },
            },
            "language": {"type": "string"},
            "payment_info": {
                "type": "dict",
                "schema": {
                    "iban": {"type": "string"},
                    "sepa_accepted": {"type": "boolean"},
                }
            },
            "suscriptions": {
                "type": "dict",
                "schema": {
                    "community_news": {"type": "boolean"},
                }
            },
            "odoo_res_users_id": {"type": "integer"},
            "odoo_res_partner_id": {"type": "integer"},
            "role": {"type": "string"},
        }
    }
}

S_MEMBER_PROFILE_RETURN_GET = {
    "member": {
        "type": "dict",
        "schema": {
            "keycloak_id": {"type": "string"},
            "name": {"type": "string"},
            "role": {"type": "string"},
            "email": {"type": "string"},
        }
    }
}

S_PROFILE_PUT = {
    "language": {"type": "string", "required": True, "empty": False}
}

S_MEMBER_PROFILE_PUT = {
    "role": {"type": "string", "required": True, "empty": False}
}

S_PROFILE_RETURN_PUT = S_PROFILE_RETURN_GET

S_MEMBER_PROFILE_RETURN_PUT = S_MEMBER_PROFILE_RETURN_GET


S_COMMUNITY_MEMBER = {
    "name": {"type": "string", "required": True, "empty": False},
    "role": {"type": "string", "required": True},
    "email": {"type": "string", "required": True},
    "keycloak_id": {"type": "string"},
    "keycloak_id": {"type": "string"},
}

S_COMMUNITY_MEMBERS_RETURN_GET = {
    "members": {
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": S_COMMUNITY_MEMBER
        }
    }
}

S_COMMUNITY_SERVICE = {
    "id": {"type": "integer"},
    "name": {"type": "string"}
}

S_COMMUNITY_RETURN_GET = {
        "community": {
        "type": "dict",
        "schema": {
            "id": {"type": "integer"},
            "name": {"type": "string"},
            "birth_date": {"type": "string"},
            "members": {
                "type": "list",
                "schema": {
                    "type": "dict",
                    "schema": S_COMMUNITY_MEMBER
                }
            },
            "contact_info": {
                "type": "dict",
                "schema": {
                    "street": {"type": "string"},
                    "postal_code": {"type": "string"},
                    "city": {"type": "string"},
                    "state": {"type": "string"},
                    "country": {"type": "string"},
                    "phone": {"type": "string"},
                    "email": {"type": "string"},
                    "telegram": {"type": "string"},
                },
            },
            "active_services": {
                "type": "list",
                "schema": {
                    "type": "dict",
                    "schema": S_COMMUNITY_SERVICE
                },
            },
            "allow_new_members": {"type": "boolean"},
        }
    }
}



from . import res_config_settings
from . import auth_oauth_provider
from . import res_company
from . import res_partner
from . import subscription_request
from . import res_users
from . import account_invoice
from . import product
from . import utm_source
from . import crm_lead
from . import keycloak_sync_wiz
from . import cm_place
from . import res_users_role



from asyncio.proactor_events import _ProactorBaseWritePipeTransport
from odoo import api, models, fields, _
from odoo.exceptions import UserError
import re


class ResPartner(models.Model):
    _inherit = 'res.partner'

    def _get_sanitized_login_username(self):
        return ((self.vat and re.sub(r"[^a-zA-Z0-9]","",self.vat).lower()) or
            (self.name and re.sub(' +','_',re.sub(r"[^a-zA-Z0-9 ]","",self.name)).lower()))

    @api.multi
    def write(self, vals):
        result = super().write(vals)
        return result

    @api.multi
    def activate_partner_in_comunitat_energetica(self):
        """ Initialize all the specific 'Comunitats Energètiques' functionallity for this partner

        It assumes that the 'res_company.id' that belongs to this partner allready exist and it is the current active company of the uid, 
        and also the corresopondent Keycloak configuration amb entities related to it are allready set in KC (kc_realm/ck_client/kc_roles/kc_groups, etc).

        Tasks:
        1) Create the related Odoo res_user.
        2) Using the KC API create the related KC user:
            - Use as username the VAT number
            - By default assign the odoo role of 'CE Member'    
        3) WIP
        
        :return: (res.user) or raise Exception
        """
        self.ensure_one()
        user = self.create_user_from_partner_id()
        if user:
            self.user_id = user
            self.sudo().push_new_user_to_keyckoack()
        else: 
            raise UserError(_("Unable to create the new Odoo user related to this partner."))
        return user

    @api.multi
    def create_user_from_partner_id(self):
        """Create the related Odoo res_user:
            - Use the VAT number as 'login'
            - By default assign the odoo role of 'CE Member'

        :return: (res.users) the new user    
        """
        user_obj = self.env["res.users"]
        login = self._get_sanitized_login_username()

        user = user_obj.search([("login", "=", login)])
        #todo: add to the found existing user the actual company_id to company_ids in case actual company is not present

        if not user:
            user = user_obj.search(
                [("login", "=", login), ("active", "=", False)]
            )
            if user:
                user.sudo().write({"active": True})
                #todo: add to the found existing user the actual company_id to company_ids in case actual company is not present
            else:
                user = user_obj.sudo().with_context(no_reset_password=True)._signup_create_user(self._get_vals_for_create_user_from_partner_id())

        return user

    @api.multi
    def _get_vals_for_create_user_from_partner_id(self):

        self.ensure_one()
        ce_roles_map = self.env['res.users'].ce_user_roles_mapping()
        ce_member_group_id = self.env['ir.model.data'].get_object_reference('ce','group_ce_member')[1]

        return {
            'email': self.email,
            'login': self._get_sanitized_login_username(),
            'partner_id': self.id,
            'company_id': self.company_id.id,
            'company_ids': [(6, 0, [self.company_id.id])],
            'groups_id': [(6,0,[9, ce_member_group_id])], # 9 = portal_user
            'role_line_ids': [(0,0,{
                'role_id': ce_roles_map['role_ce_member']['odoo_role_id'],
                'is_enabled': True,
                'company_id': self.company_id.id})],
        }

    @api.multi
    def push_new_user_to_keyckoack(self):
        self.ensure_one()
        return self.user_id.push_new_user_to_keyckoack()

    


from odoo import api, fields, models
from odoo.http import request

class ProductTemplate(models.Model):
    _inherit = "product.template"

    @api.multi
    def get_web_share_products(self, is_company):
        """We are fully overriding this function in order to make it multi-company sensitive"""
        
        if request:
            default_company = request.env["res.company"]._company_default_get()
        else:
            default_company = self.env.user_id.company_id

        target_company = default_company

        if self.env.context.get('target_odoo_company_id',False) and self.env.context.get('target_odoo_company_id') != default_company.id:
            target_company = self.env['res.company'].browse(self.env.context.get('target_odoo_company_id'))

        if is_company is True:
            product_templates = self.env["product.template"].search(
                [
                    ("is_share", "=", True),
                    ("display_on_website", "=", True),
                    ("by_company", "=", True),
                    ("company_id", "=", target_company.id),
                ]
            )
        else:
            product_templates = self.env["product.template"].search(
                [
                    ("is_share", "=", True),
                    ("display_on_website", "=", True),
                    ("by_individual", "=", True),
                    ("company_id", "=", target_company.id),
                ]
            )
        return product_templates


from email.policy import default
from odoo import api, models, fields, _
from datetime import datetime
import re
from odoo.exceptions import UserError
from odoo.addons import decimal_precision as dp
from slugify import slugify

class ResCompany(models.Model):
    _inherit = 'res.company'

    # when OVor WP are doing API calls to Odoo will use odoo_cempany_id = -1 in order to refer to the 'Coordinadora' one
    API_PARAM_ID_VALUE_FOR_COORDINADORA = -1

    # new fields
    kc_realm = fields.Char(string='KeyCloak realm name')

    coordinator = fields.Boolean(string='Platform coordinator', 
        help="Flag to indicate that this company has the rol of 'Coordinator'(=Administrator) for the current 'Comunitats Energètiques' Platform"
        )
    
    ce_admin_key_cloak_provider_id = fields.Many2one(        
        string='OAuth provider for CCEE admin',
        comodel_name='auth.oauth.provider',
    )
    auth_ce_key_cloak_provider_id = fields.Many2one(        
        string='OAuth provider for CCEE login',
        comodel_name='auth.oauth.provider',
    )

    cooperator_journal = fields.Many2one(
        "account.journal",
        string="Cooperator Journal",
        domain="[('type','=','sale'),('active','=',True)]",
        help="This journal will be"
        " the default one as the"
        " receivable journal for the"
        " cooperators"
    )

    foundation_date = fields.Date('Foundation date')
    social_telegram = fields.Char('Telegram Account')
    initial_subscription_share_amount = fields.Float('Initial Subscription Share Amount', digits=dp.get_precision('Product Price') )
    allow_new_members = fields.Boolean(string="Allow new members",default=True)

    @api.model
    def get_real_ce_company_id(self, api_param_odoo_compant_id):
        if api_param_odoo_compant_id == self.API_PARAM_ID_VALUE_FOR_COORDINADORA:
            return self.search([('coordinator','=',True)],limit=1) or None
        else:
            return self.search([('id','=',api_param_odoo_compant_id)]) or None

    @api.multi
    def check_ce_has_admin(self):
        self.ensure_one()
        admin_roles_ids = [r['odoo_role_id'] for r in self.env['res.users'].ce_user_roles_mapping().values() if r['is_admin']]
        company_user_ids = self.get_ce_members().ids
        admins_user_ids = []
        for admin_role in self.env['res.users.role'].sudo().browse(admin_roles_ids):
            for role_line in admin_role.line_ids:
                admins_user_ids.append(role_line.user_id.id)
        return any([user in admins_user_ids for user in company_user_ids])

    @api.multi
    def get_ce_members(self, domain_key='in_kc_and_active'):
        domains_dict = {'in_kc_and_active': [('company_id','=',self.id),('oauth_uid','!=',None),('active','=',True)]}
        members = self.env['res.users'].sudo().search(domains_dict['in_kc_and_active'])
        return members

    @api.model
    def _is_not_unique(self, vals):

        # check for VAT
        if vals.get('vat', False) and vals.get('vat'):
            sanit_vat = re.sub(r"[^a-zA-Z0-9]","",vals['vat']).lower()
            if sanit_vat in [re.sub(r"[^a-zA-Z0-9]","",c.vat).lower() for c in self.search([]) if c.vat]:
                raise UserError(
                    _("Unable to create new company because there is an allready existing company with this VAT number: {}").format(vals['vat']))

        # check for name
        if vals.get('name', False) and vals.get('name'):
            sanit_name = slugify(vals['name'])
            if sanit_name in [slugify(c.name) for c in self.search([]) if c.name]:
                raise UserError(
                    _("Unable to create new company because there is an allready existing company with this NAME: {}").format(vals['name']))


    @api.model
    def create(self,vals):

        # check that we are not creating duplicate companies by vat or by name
        self._is_not_unique(vals)

        new_company = super(ResCompany,self).create(vals)

        return new_company

    @api.multi
    def _community_post_company_creation_tasks(self):
        """ Do post company creation tasks that are specific for the CCEE project"""
        self.ensure_one()

        # [1] Set res.config.settings params that are company dependants
        # TODO choose the proper one depending of the legaol type of the new company
        pyme_char_template_id =  self.env['ir.model.data'].get_object_reference('l10n_es', 'account_chart_template_pymes')[1]
        #assoc_char_template_id =  self.env['ir.model.data'].get_object_reference('l10n_es', 'account_chart_template_assoc')[1]

        user_active_company_id = self.env.user.company_id.id
        user = self.env.user
        user.company_ids = [(4,self.id)]
        user.company_id = self.id
        rcs_sudo = self.env['res.config.settings'].sudo(user)
        rcs_sudo_id = rcs_sudo.create({
            'chart_template_id': pyme_char_template_id,
            'kc_realm': str(self.id),
            })
        rcs_sudo_id.execute()
        user.company_id = user_active_company_id

        # [2] Create new Sequences and Subscription Journal
        subs_acc_jour_seq = self.env['ir.sequence'].sudo().create({
            'name': 'Account Subscription Journal Sequence',
            'padding': 3,
            'prefix': 'SUBJ/%(year)s/',
            'use_date_range': True,
            'number next': 1,
            'number_increment': 1,
            'company_id': self.id,
            'date_range_ids': [(0,0,{
                'date_from':'{}-01-01'.format(datetime.now().year),
                'date_to': '{}-12-31'.format(datetime.now().year),
                'number_next':1,
                'number_next_actual':1})]
        })
        subs_acc_jour = self.env['account.journal'].sudo().create({
            'name': 'Account Subscription Journal',
            'invoice_sequence_id': subs_acc_jour_seq.id,
            'type': 'sale',
            'code': 'SUB_',
            'sequence_number next': 1,
            'company_id': self.id,
            })

        self.sudo().cooperator_journal = subs_acc_jour.id

        # [3] update several accounts
        xml_id_4400 = "{}_{}".format(self.id, 'account_common_4400') # 440000 | Deudores (euros)
        account_4400_id = self.env['ir.model.data'].sudo().get_object_reference('l10n_es', xml_id_4400)[1]

        self.sudo().property_cooperator_account = account_4400_id

        xml_id_7000 = "{}_{}".format(self.id, 'account_common_7000') # 700000 | Ventas de mercaderías en España
        account_7000_id = self.env['ir.model.data'].get_object_reference('l10n_es', xml_id_7000)[1]
        self.env.ref('easy_my_coop.product_category_company_share').sudo().property_account_income_categ_id = account_7000_id

        xml_id_600 = "{}_{}".format(self.id, 'account_common_600') # 600000 | Ventas de mercaderías en España
        account_600_id = self.env['ir.model.data'].get_object_reference('l10n_es', xml_id_600)[1]
        self.env.ref('easy_my_coop.product_category_company_share').sudo().property_account_expense_categ_id = account_600_id

        # [4] create default_share_product
        product_vals = {
            'name': 'Quota inicial alta sòcia',
            'short_name': 'Quota inicial alta sòcia',
            'sale_ok': True,
            'purchase_ok': False,
            'is_share': True,
            'display_on_website': True,
            'categ_id': self.env.ref('easy_my_coop.product_category_company_share').id,
            'type': 'service',
            'default_code': 'quota_inicial_alta_sòcia',
            'default_share_product': True,
            'list_price': self.initial_subscription_share_amount or 0.00,
            'customer': True,
            'by_company': True,
            'by_individual': True,
            'company_id': self.id,
            }
        self.env['product.template'].sudo().create(product_vals)

    @api.multi
    def _create_keycloak_realm(self):
        self.ensure_one()
        pass

    @api.multi
    def _community_post_keycloak_creation_tasks(self):
        """ Do post Kaykoac Realm creation tasks"""
        self.ensure_one()
        pass

    @api.multi
    def get_active_services(self):
        """Return a list of dicts with the key data of each active Service"""
        self.ensure_one()
        ret=[]

        # TODO: in a further iteration it will get the data from the "community services model"
        # but nowadays (2022-09-26) it still don't exists so we are getting this info from the related crm_lead.tag_ids

        creation_ce_source_id = self.env['ir.model.data'].get_object_reference('ce', 'ce_source_creation_ce_proposal')[1]
        coordinator_id = self.get_real_ce_company_id(self.API_PARAM_ID_VALUE_FOR_COORDINADORA).id

        lead_from = self.env['crm.lead'].sudo().search(
            [('company_id','=',coordinator_id),
            ('community_company_id','=',self.id),
            ('source_id','=',creation_ce_source_id)]
            ,limit=1)

        if lead_from:
            for tag in lead_from.tag_ids:
                ret.append({
                    'id': tag.id,
                    'name': tag.name,
                })
        return ret

    @api.multi
    def get_kc_groups_data(self):
        """Proceed to get the list of the KC groups related to the current company > Realm"""
        self.ensure_one()
        ce_admin_provider = self.ce_admin_key_cloak_provider_id

        if not ce_admin_provider:
            raise UserError(
                _("Unable to get the 'CE admin' provider_id related to tha current company when triying to push new user to KC."))

        wiz_vals = {
            'provider_id': ce_admin_provider.id,
            'endpoint': ce_admin_provider.users_endpoint,
            'user': ce_admin_provider.superuser,
            'pwd': ce_admin_provider.superuser_pwd,
            'login_match_key': 'username:login'
        }
        kc_wizard = self.env['auth.keycloak.sync.wiz'].create(wiz_vals)
        kc_wizard._validate_setup()
        token = kc_wizard._get_token()

        return kc_wizard._get_realm_groups_data(token)


import logging
from odoo.addons.component.core import Component
from odoo.addons.base_rest.http import wrapJsonException
from werkzeug.exceptions import BadRequest
from odoo import _
from . import schemas

_logger = logging.getLogger(__name__)


class CRMLeadService(Component):
    _inherit = "base.rest.private_abstract_service"
    _name = "crm.lead.services"
    _collection = "ce.services"
    _usage = "crm-lead"
    _description = """
        CRMLead requests
    """

    def create(self, **params):
        company_id = self.env['res.company'].get_real_ce_company_id(params['odoo_company_id']).id
        params.update({'odoo_company_id': company_id})

        sources = {s.name:s.res_id for s in self.env['ir.model.data'].search([
            ('module','=','ce'), ('model','=','utm.source')])}
        if params['source_xml_id'] not in sources:
            raise wrapJsonException(
                BadRequest(
                    _("Source {} not found").format(
                        params["source_xml_id"])
                ),
                include_description=True,
            )

        target_source_xml_id = params['source_xml_id']
        params.update({'source_xml_id': sources[params['source_xml_id']]})

        try:
            tag_ids = params['tag_ids']
        except:
            tag_ids = False
        if tag_ids:
            for tag_id in tag_ids:
                tag_id_res = self.env['crm.lead.tag'].search([('id','=',tag_id)]).id
                if not tag_id_res:
                    raise wrapJsonException(
                        BadRequest(
                            _("Tag {} not found").format(tag_id)
                        ),
                        include_description=True,
                    )

        params = self._prepare_create(params)
        sr = self.env["crm.lead"].sudo().create(params)

        # send the receipt confirmation email
        email_template = None
        if target_source_xml_id == 'ce_source_creation_ce_proposal':
            email_template = 'email_templ_lead_ce_creation_receipt_confirm_id'
        elif target_source_xml_id == 'ce_source_existing_ce_contact':
            email_template = 'email_templ_lead_request_contact_confirm_id'
        elif target_source_xml_id == 'ce_source_existing_ce_info':
            email_template = 'email_templ_lead_request_ce_news_confirm_id'
        elif target_source_xml_id == 'ce_source_future_location_ce_info':
            email_template = 'email_templ_lead_request_advise_future_ce_confirm_id'
        elif target_source_xml_id == 'ce_source_general_info':
            email_template = 'email_templ_lead_request_platform_news_confirm_id'

        if email_template:
            self.env.ref('ce.{}'.format(email_template)).sudo().send_mail(sr.id)

        return self._to_dict(sr)

    def _validator_create(self):
        source = self.work.request.params.get('source_xml_id', False)
        if source == 'ce_source_creation_ce_proposal':
            return schemas.S_CRM_LEAD_CREATE_ALTA_CE
        return schemas.S_CRM_LEAD_CREATE

    def _validator_return_create(self):
        return schemas.S_CRM_LEAD_RETURN_CREATE

    @staticmethod
    def _to_dict(crm_lead):
        return {
            "id": crm_lead.id
        }

    def _prepare_create(self, params):

        contact_name = params.get("partner_firstname", None) or params.get("partner_lastname",None) or None
        if params.get("partner_firstname",None) and params.get("partner_lastname",None):
            contact_name =  "{} {}".format(params.get("partner_firstname",""),params.get("partner_lastname",""))

        target_lang = None
        if params.get("partner_language", None):
            active_langs = self.env['res.lang'].search([('active','=', True)])
            active_code_langs = [l.code.split('_')[0] for l in active_langs]
            if params.get('partner_language').lower() not in active_code_langs:
                raise wrapJsonException(
                    BadRequest(),
                    include_description=False,
                    extra_info={'message': _("This language code %s is not active in Odoo. Active ones: %s") % (
                        params.get('partner_language').lower(),
                        str(active_code_langs))}
                )
            target_lang = [l for l in active_langs if l.code.split('_')[0] == params.get('partner_language').lower()][0]

        vals = {
            "submission_type": 'place_proposal_submission',
            "type": 'opportunity',
            "contact_name": contact_name,
            "name": params.get("partner_name"),
            "description": params.get("partner_description"),
            "street": params.get("partner_full_address"),
            "zip": params.get("partner_zip"),
            "city": params.get("partner_city"),
            "email_from": params.get("partner_email"),
            "phone": params.get("partner_phone"),
            "tag_ids": [(6, 0, params.get("tag_ids", []))],
            "company_id": params.get("odoo_company_id"),
            "source_id": params.get("source_xml_id"),
            "form_submission_metadata_ids": [
                (0,0,{'key':'partner_qty_members','value':params.get("partner_qty_members"),'type':'string'}),
                (0,0,{'key':'partner_legal_state','value':params.get("partner_legal_state"),'type':'string'}),
                (0,0,{'key':'partner_foundation_date','value':params.get("partner_foundation_date"),'type':'string'}),
                (0,0,{'key':'partner_vat','value':params.get("partner_vat"),'type':'string'}),
                (0,0,{'key':'partner_comments','value':params.get("partner_comments"),'type':'string'}),
                (0,0,{'key':'partner_state','value':params.get("partner_state"),'type':'string'}),
                (0,0,{'key':'partner_firstname','value':params.get("partner_firstname"),'type':'string'}),
                (0,0,{'key':'partner_lastname','value':params.get("partner_lastname"),'type':'string'}),
                (0,0,{'key':'contact_email','value':params.get("partner_email"),'type':'string'}),
                (0,0,{'key':'contact2_firstname','value':params.get("contact2_firstname"),'type':'string'}),
                (0,0,{'key':'contact2_lastname','value':params.get("contact2_lastname"),'type':'string'}),
                (0,0,{'key':'contact2_email','value':params.get("contact2_email"),'type':'string'}),
                (0,0,{'key':'contact2_mobile','value':params.get("contact2_mobile"),'type':'string'}),
                (0,0,{'key':'partner_map_place_form_url','value':params.get("partner_map_place_form_url"),'type':'string'}),
                (0,0,{'key':'partner_language','value':target_lang and target_lang.code or 'en_US','type':'string'}),
                (0,0,{'key':'partner_twitter','value':'','type':'string'}),
                (0,0,{'key':'partner_facebook','value':'','type':'string'}),
                (0,0,{'key':'partner_instagram','value':'','type':'string'}),
                (0,0,{'key':'partner_telegram','value':'','type':'string'}),
                (0,0,{'key':'partner_group_image_url','value':'','type':'string'}),
                (0,0,{'key':'partner_latitude','value':'','type':'string'}),
                (0,0,{'key':'partner_longitude','value':'','type':'string'}),
                (0,0,{'key':'partner_initial_share_amount','value':'100','type':'string'})
                ],
        }
        if params.get("partner_phone", False) and params.get("partner_phone").strip()[:1] in ('6','7'):
            vals['mobile'] = params.get("partner_phone")

        return vals



from odoo import api, models, fields, _

class CmPlace(models.Model):
    _inherit = 'crm.team'

    community_company_id = fields.Many2one(        
        string='Related Community',
        comodel_name='res.company',
        domain="[('coordinator','!=',True)]",
        help="Community related to this Map point"
    )
 


from odoo.addons.base_rest.controllers import main


class MainController(main.RestController):
    _root_path = "/api/"
    _collection_name = "ce.services"
    _default_auth = "api_key"


class CrmLeadController(MainController):
    pass



from odoo import api, models, fields, _
from odoo.exceptions import UserError

class AccountInvoice(models.Model):
    _inherit = 'account.invoice'

    def create_user(self, partner):
        """Caution: we are full overriding this function in order to avoid the
        call to action_reset_password that will send an email to the user from odoo

        :return: (res.users) the new user
        """
        return partner.activate_partner_in_comunitat_energetica()


from odoo import models, fields


class UtmSource(models.Model):
    _inherit = 'utm.source'

    source_ext_id = fields.Char('ID Ext Source', compute='compute_ext_id_source')

    def compute_ext_id_source(self):
        for record in self:
            res = record.get_external_id()
            record.source_ext_id = False
            if res.get(record.id):
                record.source_ext_id = res.get(record.id)


import logging
import json
from odoo.addons.base_rest import restapi
from werkzeug.exceptions import BadRequest, NotFound
from odoo.addons.base_rest.http import wrapJsonException
from odoo.addons.component.core import Component
from odoo import _
from datetime import datetime
from . import schemas

_logger = logging.getLogger(__name__)


class CommunityService(Component):
    _inherit = "base.rest.private_abstract_service"
    _name = "ce.community.services"
    _collection = "ce.services"
    _usage = "community"
    _description = """
        CE community requests
    """

    @restapi.method(
        [(["/<int:odoo_company_id>", "/<int:odoo_company_id>/<string:method_name>"], "GET")],
        output_param=restapi.CerberusValidator(
            "_validator_return_community_get"),
        auth="api_key",
    )
    def get(self, _odoo_company_id):
        endpoint_args = self.work.request.endpoint_arguments
        company_id = self.env['res.company'].get_real_ce_company_id(
            _odoo_company_id)

        if not company_id:
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={'message': _(
                    "No Odoo Company found for odoo_company_id %s") % _odoo_company_id}
            )

        if 'method_name' in endpoint_args and endpoint_args.get('method_name') == 'members':
            return self._to_dict_members(company_id.get_ce_members())
        else:
            return self._to_dict_community(company_id)

    def _validator_return_community_get(self):
        endpoint_args = self.work.request.endpoint_arguments
        if 'method_name' in endpoint_args and endpoint_args.get('method_name') == 'members':
            return schemas.S_COMMUNITY_MEMBERS_RETURN_GET
        else:
            return schemas.S_COMMUNITY_RETURN_GET

    @staticmethod
    def _to_dict_members(users):
        resp = {'members': []}
        for user in users:
            resp['members'].append({
                "name": '{} {}'.format(user.firstname, user.lastname,),
                "role": user.ce_role or "",
                "email": user.email or "",
                "keycloak_id": user.oauth_uid,
            })
        return resp

    @staticmethod
    def _to_dict_community(company):

        resp = {'community':{
            'id': company.id,
            'name': company.name,
            'birth_date': company.foundation_date and company.foundation_date.strftime('%Y-%m-%d') or '',
            'members': [],
            'contact_info': {
                'street': company.street or '',
                'postal_code': company.zip or '',
                'city': company.city or '',
                'state': company.state_id.name or '',
                'country': company.country_id.name or '',
                'phone': company.phone or '',
                'email': company.email or '',
                'telegram': company.social_telegram or ''},
            'active_services': [],
            'allow_new_members': company.allow_new_members,
        }}

        resp['community'].update(CommunityService._to_dict_members(
            company.get_ce_members()))
        resp['community'].update({'active_services': company.get_active_services()})

        return resp



{
    'name': "Comunitats Energètiques customizations",
    'version': '12.0.0.0.10',
    'depends': [
        'base_rest_base_structure',
        'cooperator_website',
        'auth_keycloak',
        'contacts',
        'base_rest',
        'base_user_role',
        'auth_api_key',
        'crm',
        'community_maps',
    ],
    'author': "Coopdevs Treball SCCL & Som Energia SCCL",
    'website': 'https://somenergia.coop',
    'category': "Cooperative management",
    'description': """
    Odoo Comunitats Energètiques customizations.
    """,
    "license": "AGPL-3",
    'demo': [
        'demo/demo_data.xml',
    ],
    'data': [
        'security/ir_rules_data.xml',
        'data/res_groups_data.xml',
        'data/auth_keycloak_data.xml',
        'data/ir_config_parameter_data.xml',
        'data/res_company_data.xml',
        'data/utm_data.xml',
        'data/crm_lead_tag.xml',
        'data/crm_maps_data.xml',
        'views/res_config_settings.xml',
        'views/crm_lead_views.xml',
        'views/auth_oauth_provider_views.xml',
        'views/res_company_views.xml',
        'views/website_subscription_template.xml',
        'views/ce_views.xml',
        'views/utm_views.xml',
        'views/menus.xml',
        'views/res_users.xml',
        'views/cm_place.xml',
        'data/mail_template_data.xml',
        'data/mail_template_update_data.xml'
    ],
    'installable': True,
    'application': True,
    'auto_install': False
}





from odoo import http
from odoo.http import request
from odoo.tools.translate import _
from odoo.addons.cooperator_website.controllers import main as emyc_wsc


class WebsiteSubscriptionCCEE(emyc_wsc.WebsiteSubscription):
    @http.route()
    def display_become_cooperator_page(self, **kwargs):

        target_odoo_company_id = False
        if kwargs.get('odoo_company_id', False):
            try:
                target_odoo_company_id = int(kwargs.get('odoo_company_id'))
            except:
                pass

        if ('odoo_company_id' in kwargs) and (not target_odoo_company_id or not request.env['res.company'].sudo().search(
                [('id', '=', target_odoo_company_id)])):
            return http.Response(_("Not valid parameter value [odoo_company_id]"), status=500)

        ctx = dict(request.context)
        ctx.update({'target_odoo_company_id': target_odoo_company_id})
        request.context = ctx


        res = super(WebsiteSubscriptionCCEE,
                    self).display_become_cooperator_page(**kwargs)
        return res

    def fill_values(self, values, is_company, logged, load_from_user=False):
        values = super(WebsiteSubscriptionCCEE, self).fill_values(
            values, is_company, logged, load_from_user)

        default_company = request.env["res.company"]._company_default_get()


        target_company_id = request.context.get('target_odoo_company_id', False) and int(
            request.context.get('target_odoo_company_id')) or None
        if target_company_id and target_company_id != default_company.id:
            company = request.env['res.company'].sudo().search(
                [('id', '=', target_company_id)])
    
            values["company"] = company
            if not values.get("country_id"):
                if company.default_country_id:
                    # company.default_country_id.id
                    values["country_id"] = "68"
                else:
                    values["country_id"] = "68"
            if not values.get("activities_country_id"):
                if company.default_country_id:
                    # company.default_country_id.id
                    values["activities_country_id"] = "68"
                else:
                    values["activities_country_id"] = "68"
            if not values.get("lang"):
                if company.default_lang_id:
                    values["lang"] = company.default_lang_id.code

            values.update(
                {
                    "display_data_policy": company.display_data_policy_approval,
                    "data_policy_required": company.data_policy_approval_required,
                    "data_policy_text": company.data_policy_approval_text,
                    "display_internal_rules": company.display_internal_rules_approval,
                    "internal_rules_required": company.internal_rules_approval_required,
                    "internal_rules_text": company.internal_rules_approval_text,
                    "display_financial_risk": company.display_financial_risk_approval,
                    "financial_risk_required": company.financial_risk_approval_required,
                    "financial_risk_text": company.financial_risk_approval_text,
                    "display_generic_rules": company.display_generic_rules_approval,
                    "generic_rules_required": company.generic_rules_approval_required,
                    "generic_rules_text": company.generic_rules_approval_text,
                }
            )

        return values




import logging, json
from odoo.addons.base_rest import restapi
from werkzeug.exceptions import BadRequest, NotFound
from odoo.addons.base_rest.http import wrapJsonException
from odoo.addons.component.core import Component
from odoo import _
from . import schemas

_logger = logging.getLogger(__name__)


class MemberProfileService(Component):
    _inherit = "base.rest.private_abstract_service"
    _name = "ce.member.profile.services"
    _collection = "ce.services"
    _usage = "profile"
    _description = """
        CE Member profile requests
    """
    @restapi.method(
        [(["/<string:keycloak_id>"], "GET")],
        output_param=restapi.CerberusValidator("_validator_return_get"),
        auth="api_key",
    )
    def get(self, _keycloak_id):
        user, partner, partner_bank, sepa_mandate = self._get_profile_objs(_keycloak_id)
        return self._to_dict(user, partner, partner_bank, sepa_mandate)
    
    def _validator_return_get(self):
        return schemas.S_PROFILE_RETURN_GET

    @restapi.method(
        [(["/<string:keycloak_id>"], "PUT")],
        input_param=restapi.CerberusValidator("_validator_update"),
        output_param=restapi.CerberusValidator("_validator_return_update"),
        auth="api_key",
    )
    def update(self, _keycloak_id, **params):
        user, partner, partner_bank, sepa_mandate = self._get_profile_objs(_keycloak_id)
        active_langs = self.env['res.lang'].search([('active','=', True)])
        active_code_langs = [l.code.split('_')[0] for l in active_langs]

        if params.get('language').lower() not in active_code_langs:
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={'message': _("This language code %s is not active in Odoo. Active ones: %s") % (
                    params.get('language').lower(),
                    str(active_code_langs))}
            )

        target_lang = [l for l in active_langs if l.code.split('_')[0] == params.get('language').lower()][0]
        if partner.lang != target_lang.code:
            partner.sudo().write({'lang':target_lang.code})

        #also update lang in KeyCloack related user throw API call
        try:
            user.update_user_data_to_keyckoack(['lang'])
        except Exception as ex:
            details = (ex and hasattr(ex, 'name') and " Details: {}".format(ex.name)) or ''
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={
                    'message': _("Unable to update the lang in Keycloak for the related KC user ID: {}.{}").format(
                        user.oauth_uid,
                        details),
                    'code': 500,
                })

        return self._to_dict(user, partner, partner_bank, sepa_mandate)

    def _validator_update(self):
        return schemas.S_PROFILE_PUT

    def _validator_return_update(self):
        return schemas.S_PROFILE_RETURN_PUT

    @staticmethod
    def _to_dict(user, partner, bank, sepa_mandate):

        return {'profile':{
            "keycloak_id": user.oauth_uid,
            "odoo_res_users_id": user.id,
            "odoo_res_partner_id": user.partner_id.id,
            "name": user.firstname,
            "surname": user.lastname,
            "birth_date": partner.birthdate_date and partner.birthdate_date.strftime("%Y-%m-%d") or "",
            "gender": partner.gender or "",
            "vat": partner.vat or "",
            "contact_info": {
                "email": partner.email or "",
                "phone": partner.mobile or partner.phone or "",
                "street": partner.street or "",
                "postal_code": partner.zip or "",
                "city": partner.city or "",
                "state": partner.state_id and partner.state_id.name or "",
                "country": partner.country_id and partner.country_id.name or ""
            },
            "language": partner.lang and partner.lang.split('_')[0] or "",
            "payment_info": {
                "iban": bank and bank.acc_number or "",
                "sepa_accepted": sepa_mandate
            },
            "role": user.ce_role or "",
        }}

    def _get_profile_objs(self, _keycloak_id):
        user = self.env["res.users"].sudo().search([('oauth_uid','=',_keycloak_id)])
        if not user:
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={'message': _("No Odoo User found for KeyCloak user id %s") % _keycloak_id}
            )

        partner = user.partner_id or None
        if not partner:
            raise wrapJsonException(
                BadRequest(),
                include_description=False,
                extra_info={'message': _("No Odoo Partner found for Odoo user with login username %s") % user.login}
            )

        partner_bank = self.env['res.partner.bank'].sudo().search([
                ('partner_id', '=', partner.id), ('company_id', '=', partner.company_id.id)
            ], order="sequence asc", limit=1) or None

        sepa_mandate = partner_bank and any([sm.id for sm in partner_bank.mandate_ids if sm.state == 'valid']) or False

        return user, partner, partner_bank, sepa_mandate
